;;; rx-to-peg.el --- Transform simple rx forms into really equivalent pegs -*- lexical-binding: t -*-

;; Copyright (C) 2021 Free Software Foundation, Inc

;; For a peg implementation in Emacs see peg.el in Gnu Elpa.

;; The algorithm is from here:
;;   http://www.lua.inf.puc-rio.br/publications/medeiros11regular.pdf
;;
;; We use the terms "peg" and "grammar" synonymously.  A peg is
;; represented as pair (i.e. list) (P p) where P the list of rule
;; definitions and p the expression to match.
;;
;; Use like this:
;;
;; (rx->peg '(and (* (and (or "a" "aa") "b")) eos))
;;   ==>
;; (((a (or (and "ab"  a)
;;          (and "aab" a)
;;          (eob))))
;;  a)

;;; Code:


;; Helpers for handling grammars.

;; these two help to construct nicer `and' and `or' expressions

(defun g-or (x y)
  (let ((conds (append (if (eq (car-safe x) 'or) (cdr x) (list x))
                       (if (eq (car-safe y) 'or) (cdr y) (list y)))))
    (if (cdr conds)
        `(or ,@conds)
      (car conds))))

(defun g-and (x y)
  (cl-flet ((concat-subsequent-strings-in (list)
              (let ((rest list))
                (while rest
                  (when (stringp (car rest))
                    (while (stringp (cadr rest))
                      (setcar rest (concat (car rest) (cadr rest)))
                      (setcdr rest (cddr rest))))
                  (cl-callf cdr rest)))
              list))
    (let ((conds (delete "" (concat-subsequent-strings-in
                             (append (if (eq (car-safe x) 'and) (cdr x) (list x))
                                     (if (eq (car-safe y) 'and) (cdr y) (list y)))))))
      (if (cdr conds)
          `(and ,@conds)
        (car conds)))))


(defvar peg-symbol-ctr)

(defun g--get-rule-name (n)
  "Return non-terminal (symbol) number N."
  (intern
   (if (<= n 26)
       (string (+ (- ?a 1) n))
     (format "r%d" n))))

(defun g-make (P p) (list P p))

(defun g-rules (g)
  ;; Return G's rule defintions
  (car g))

(defun g-expression (g)
  ;; Return G's match expression
  (cadr g))

(defun g-w/extended-expr (g f)
  ;; apply F to G's match expression and return the result as new peg
  (pcase-let ((`(,P ,p) g))
    `(,P ,(funcall f p))))


;; Definition of the transcription function Π as in the paper

(cl-defgeneric Π (expr cont)
  ;; EXPR is a regular expression (as an s-exp).
  ;; Second arg CONT is the continuation grammar.
  ;;
  ;; When the following two conditions hold:
  ;;
  ;;   (1) For no subrx (* E) of EXPR does E match the empty string.
  ;;   (2) The language of EXPR has the prefix property: There are no
  ;;   distinct string S1 and S2 in the language of EXPR so that S1 is a
  ;;   prefix of S2
  ;;
  ;; then the returned peg is equivalent to EXPR, i.e. both match the
  ;; same strings.  (2) can be ensured by matching the end of the string
  ;; (`eos' in `rx')
  )

;; Our entry point:
(defun rx->peg (rx)
  (let ((peg-symbol-ctr 0))
    ;; call Π with rx and a continuation grammar with language {""}
    (Π rx '(() ""))))

(cl-defmethod Π ((expr string) cont)
  (g-w/extended-expr cont (lambda (p) (g-and expr p))))

(cl-defmethod Π ((_expr (eql eos)) cont)
  ;; pegs search buffers, so eos gets eob
  (g-w/extended-expr cont (lambda (p) (g-and '(eob) p))))

(cl-defmethod Π ((expr (head and)) cont)
  (cl-reduce #'Π (cdr expr) :from-end t :initial-value cont))

(cl-defmethod Π ((expr (head or)) cont)
  (pcase (cdr expr)
    ((and (pred cddr) factors)
     (Π `(or ,(car factors) (or ,@(cdr factors))) cont))
    ((and (pred cdr) `(,e1 ,e2))
     (let* ((g1 (Π e1 cont))
            (g2 (Π e2 (g-w/extended-expr g1 (lambda (_) (g-expression cont))))))
       (g-w/extended-expr g2 (lambda (p2) (g-or (g-expression g1) p2)))))
    (ex (Π ex cont))))

(cl-defmethod Π ((expr (head *)) cont)
  (let* ((e (cadr expr))
         (rulesym (g--get-rule-name (cl-incf peg-symbol-ctr)))
         (g1 (Π e (g-make (g-rules cont) rulesym))))
    (g-make (cons (list rulesym (g-or (g-expression g1)
                                      (g-expression cont)))
                  (g-rules g1))
            rulesym)))

(provide 'rx-to-peg)
;;; Π.el ends here
